/*************************************************************************
 *  playmus.c
 *
 *  Copyright (C) 1995 Michael Heasley (mheasley@hmc.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include "musserver.h"

void seqbuf_dump(void);
void reset_midi(void);
void note_off(int note, int channel, int volume);
void note_on(int note, int channel, int volume);
void pitch_bend(int channel, int value);
void control_change(int controller, int channel, int value);
void patch_change(int patch, int channel);
void midi_wait(float time);
void midi_timer(int action);

#define START 0
#define STOP 1
#define CONT 2

int returnval = 0;


void next_level()
{
  reset_midi();
  returnval = 1;
}

void prev_level()
{
  reset_midi();
  returnval = -1;
}

void secret_level()
{
  reset_midi();
  returnval = 2;
}

void quitmus()
{
  reset_midi();
  returnval = 3;
}


int playmus(char *musdata, unsigned int mussize, int play_once)
{
unsigned char event;
unsigned char event1;
unsigned char event2;
unsigned char eventtype;
unsigned char channelnum;
unsigned char last = 0;
unsigned char notenum;
unsigned char notevol;
unsigned int pitchwheel;
unsigned char controlnum;
unsigned char controlval;
unsigned int ticks;
unsigned char tmpch;
unsigned int done = 0;
unsigned int lastvol[16];
double delaytime;
double curtime = 0.0;
char *muscopy;


signal(SIGHUP, next_level);
signal(SIGUSR1, prev_level);
signal(SIGUSR2, secret_level);
signal(SIGQUIT, quitmus);
signal(SIGINT, quitmus);
signal(SIGTERM, quitmus);
signal(SIGCONT, SIG_IGN);

returnval = 0;
muscopy = musdata;
midi_timer(START);

while(!done) {
  if (returnval != 0)
    return returnval;
  event = *musdata++;
  channelnum = event & 15;
  if (channelnum > 8)
    channelnum++;
  if (channelnum == 16)
    channelnum = 9;
  eventtype = (event >> 4) & 7;
  last = event & 128;
#ifdef DEBUG
  printf("\nEvent byte: %d\n", event);
  printf("Channel #:  %d\n", channelnum);
  printf("Event type: %d\n", eventtype);
  printf("Last:       %d\n", last);
#endif

  switch (eventtype)
    {
    case 0:		/* note off */
      event1 = *musdata++;
      notenum = event1 & 127;
      note_off(notenum, channelnum, lastvol[channelnum]);
      break;

    case 1:		/* note on */
      event1 = *musdata++;
      notenum = event1 & 127;
#ifdef DEBUG
      printf("\nEvent: note on\n");
      printf("Note number: %d\n", notenum);
#endif
      if (event1 & 128)
        {
        event2 = *musdata++;
        notevol = event2 & 127;
        lastvol[channelnum] = notevol;
#ifdef DEBUG
        printf("Note volume: %d\n", notevol);
#endif
        }
      note_on(notenum, channelnum, lastvol[channelnum]);
      break;

    case 2:		/* pitch wheel */
      event1 = *musdata++;
      pitchwheel = event1 / 2;
      pitch_bend(channelnum, pitchwheel);
#ifdef DEBUG
      printf("\nEvent: pitchwheel\n");
      printf("Pitchwheel value: %d\n", pitchwheel);
#endif
      break;

    case 4:		/* midi controller change */
      event1 = *musdata++;
      controlnum = event1 & 127;
      event2 = *musdata++;
      controlval = event2 & 127;
#ifdef DEBUG
      printf("\nEvent: controller change\n");
      printf("Controller number: %d\n", controlnum);
      printf("Controller value: %d\n", controlval);
#endif
      switch (controlnum)
        {
        case 0:		/* patch change */
          patch_change(controlval, channelnum);
          break;
        case 3:		/* volume */
          control_change(CTL_MAIN_VOLUME, channelnum, controlval);
          break;
        case 4:		/* pan */
          control_change(CTL_PAN, channelnum, controlval);
          break;
        }
      break;

    case 6:	/* end of music data */
        if (play_once)
          {
          returnval = 1;
          pause();
          }
        else
          {
          musdata = muscopy;
          midi_timer(START);
          curtime = 0.0;
          last = 0;
          }
#ifdef DEBUG
      printf("\nEnd of music data\n");
#endif
      break;

    case 3:	/* unknown, but contains data */
      musdata++;
      break;

    default:	/* unknown */
      break;
    }
  if (last)	/* next data portion is time data */
    {
    tmpch = *musdata++;
    ticks = tmpch & 127;
    while(tmpch & 128)
      {
      tmpch = *musdata++;
      ticks = (ticks * 128) + (tmpch & 127);
      }
    delaytime = (100.0 * (double)ticks) / 128.0;
    curtime += delaytime;
#ifdef DEBUG
    printf("\nTicks to next event: %d  (delaytime = %ld\n", ticks, delaytime);
#endif
    midi_wait(curtime);
    }
  }
  return returnval;
}
